from __future__ import annotations

from typing import Any

import pytest

import ckan.plugins.toolkit as tk
from ckan import types
from ckan.tests.helpers import call_action


@pytest.fixture(autouse=True)
def prepare(reset_redis: Any):
    reset_redis()


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestPermissionManage:
    def test_anonymous_is_not_allowed(self):
        """Anonymous users are not allowed to manage files."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access("files_permission_manage_files", {"user": ""}, {})

    def test_authenticated_is_not_allowed(self, user: dict[str, Any]):
        """Authenticated users are not allowed to manage files."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access("files_permission_manage_files", {"user": user["name"]}, {})


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestPermissionOwns:
    def test_anonymous_is_not_allowed(self, file: dict[str, Any]):
        """Anonymous users is not an owner."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access("files_permission_owns_file", {"user": ""}, {"id": file["id"]})

    def test_authenticated_is_not_allowed(self, user: dict[str, Any], file: dict[str, Any]):
        """Authenticated users is not an owner."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access("files_permission_owns_file", {"user": user["name"]}, {"id": file["id"]})

    def test_owner_is_allowed(self, user: dict[str, Any], file_factory: types.TestFactory):
        """Owners users is identified."""
        file = file_factory(user=user)
        tk.check_access("files_permission_owns_file", {"user": user["name"]}, {"id": file["id"]})


@pytest.mark.usefixtures("with_plugins", "clean_db")
class BasePermission:
    action: str

    def test_anonymous_is_not_allowed(self, file: dict[str, Any]):
        """Anonymous user does not have permission."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access(self.action, {"user": ""}, {"id": file["id"]})

    def test_authenticated_is_not_allowed(self, user: dict[str, Any], file: dict[str, Any]):
        """Authenticated user does not have permission."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access(self.action, {"user": user["name"]}, {"id": file["id"]})

    def test_owner_is_allowed(self, user: dict[str, Any], file_factory: types.TestFactory):
        """File owner has permission for owned file."""
        file = file_factory(user=user)
        tk.check_access(self.action, {"user": user["name"]}, {"id": file["id"]})

    @pytest.mark.ckan_config("ckanext.files.owner.cascade_access", {})
    def test_owner_without_cascade_is_not_allowed(
        self,
        user: dict[str, Any],
        file_factory: types.TestFactory,
        package_factory: types.TestFactory,
    ):
        """Entity owners do not have permissions for file owned by entity without cascade."""
        package = package_factory(user=user)
        file = file_factory(user=user)
        call_action(
            "files_transfer_ownership",
            id=file["id"],
            owner_type="package",
            owner_id=package["id"],
        )
        with pytest.raises(tk.NotAuthorized):
            tk.check_access(self.action, {"user": user["name"]}, {"id": file["id"]})

    def test_owner_with_cascade_is_allowed(
        self,
        user: dict[str, Any],
        file_factory: types.TestFactory,
        package_factory: types.TestFactory,
    ):
        """Entity owners have permissions for file owned by entity with cascade."""
        package = package_factory(user=user)
        file = file_factory(user=user)
        call_action(
            "files_transfer_ownership",
            id=file["id"],
            owner_type="package",
            owner_id=package["id"],
        )
        tk.check_access(self.action, {"user": user["name"]}, {"id": file["id"]})


class TestPermissionEdit(BasePermission):
    action = "files_permission_edit_file"


class TestPermissionDelete(BasePermission):
    action = "files_permission_delete_file"


class TestPermissionRead(BasePermission):
    action = "files_permission_read_file"


class TestPermissionDownload(BasePermission):
    action = "files_permission_download_file"


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileCreate:
    def test_anonymous_uploads_are_not_allowed(self):
        """Anonymous users are not allowed to upload files."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access("files_file_create", {"user": ""}, {"storage": "test"})

    def test_authenticated_uploads_are_not_allowed_by_default(self, user: dict[str, Any]):
        """Authenticated users are not allowed to upload files with default settings."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access("files_file_create", {"user": user["name"]}, {"storage": "test"})

    @pytest.mark.ckan_config("ckanext.files.authenticated_uploads.allow", True)
    @pytest.mark.ckan_config("ckanext.files.authenticated_uploads.storages", ["test"])
    def test_authenticated_uploads_can_be_enabled(self, user: dict[str, Any]):
        """Authenticated users can upload files if corresponding option is enabled."""
        tk.check_access("files_file_create", {"user": user["name"]}, {"storage": "test"})


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileTrack:
    def test_anonymous_are_not_allowed(self):
        """Anonymous users are not allowed to register files."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access("files_file_register", {"user": ""}, {"storage": "test"})

    def test_authenticated_are_not_allowed(self, user: dict[str, Any]):
        """Authenticated users are not allowed to register files."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access("files_file_register", {"user": user["name"]}, {"storage": "test"})


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestFileSearch:
    def test_anonymous_search_is_not_allowed(self):
        """Anonymous users are not allowed to search files."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access("files_file_create", {"user": ""}, {"storage": "test"})

    def test_authenticated_search_is_not_allowed(self, user: dict[str, Any]):
        """Authenticated users are not allowed to search files."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access("files_file_create", {"user": user["name"]}, {"storage": "test"})


@pytest.mark.usefixtures("with_plugins", "clean_db")
class BaseOperation:
    action: str

    def test_anonymous_is_not_allowed(self, file: dict[str, Any]):
        """Anonymous users are not allowed to perform operation."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access(self.action, {"user": ""}, {"id": file["id"]})

    def test_authenticated_is_not_allowed(self, user: dict[str, Any], file: dict[str, Any]):
        """Authenticated users are not allowed to perform operation."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access(self.action, {"user": user["name"]}, {"id": file["id"]})

    def test_owner_is_allowed(self, user: dict[str, Any], file_factory: types.TestFactory):
        """Owners users are allowed to perform operation."""
        file = file_factory(user=user)
        tk.check_access(self.action, {"user": user["name"]}, {"id": file["id"]})


class TestFileDelete(BaseOperation):
    action = "files_file_delete"


class TestFileShow(BaseOperation):
    action = "files_file_show"


class TestFileRename(BaseOperation):
    action = "files_file_rename"


class TestFilePin(BaseOperation):
    action = "files_file_pin"


class TestFileUnpin(BaseOperation):
    action = "files_file_unpin"


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestOwnershipTransfer:
    def test_anonymous_transfer_is_not_allowed(self, file: dict[str, Any]):
        """Anonymous users are not allowed to transfer files."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access(
                "files_transfer_ownership",
                {"user": ""},
                {
                    "id": file["id"],
                    "owner_id": file["owner_id"],
                    "owner_type": file["owner_type"],
                },
            )

    def test_authenticated_transfer_is_not_allowed(self, user: dict[str, Any], file: dict[str, Any]):
        """Authenticated users are not allowed to transfer files."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access(
                "files_transfer_ownership",
                {"user": user["name"]},
                {
                    "id": file["id"],
                    "owner_id": file["owner_id"],
                    "owner_type": file["owner_type"],
                },
            )

    def test_owner_transfer_is_allowed_when_unpinned(self, user: dict[str, Any], file_factory: types.TestFactory):
        """Owners users are not allowed to transfer pinned files."""
        file = file_factory(user=user)

        tk.check_access(
            "files_transfer_ownership",
            {"user": user["name"]},
            {
                "id": file["id"],
                "owner_id": file["owner_id"],
                "owner_type": file["owner_type"],
                "force": False,
            },
        )

        call_action("files_file_pin", id=file["id"])
        with pytest.raises(tk.NotAuthorized):
            tk.check_access(
                "files_transfer_ownership",
                {"user": user["name"]},
                {
                    "id": file["id"],
                    "owner_id": file["owner_id"],
                    "owner_type": file["owner_type"],
                    "force": False,
                },
            )

        tk.check_access(
            "files_transfer_ownership",
            {"user": user["name"]},
            {
                "id": file["id"],
                "owner_id": file["owner_id"],
                "owner_type": file["owner_type"],
                "force": True,
            },
        )

    @pytest.mark.ckan_config("ckanext.files.owner.transfer_as_update", False)
    def test_owner_transparent_transfer(
        self,
        user: dict[str, Any],
        file_factory: types.TestFactory,
        package_factory: types.TestFactory,
    ):
        """Transfer does not work without transfer-as-update because of missing auth function."""
        package = package_factory(user=user)
        file = file_factory(user=user)
        with pytest.raises(ValueError):  # noqa: PT011
            tk.check_access(
                "files_transfer_ownership",
                {"user": user["name"]},
                {
                    "id": file["id"],
                    "owner_id": package["id"],
                    "owner_type": "package",
                    "force": False,
                },
            )

    def test_owner_transfer_as_update(
        self,
        user: dict[str, Any],
        file_factory: types.TestFactory,
        package_factory: types.TestFactory,
    ):
        """Enabled transfer-as-update allows transfer without additional code."""
        package = package_factory(user=user)
        file = file_factory(user=user)
        tk.check_access(
            "files_transfer_ownership",
            {"user": user["name"]},
            {
                "id": file["id"],
                "owner_id": package["id"],
                "owner_type": "package",
                "force": False,
            },
        )


@pytest.mark.usefixtures(
    "with_plugins",
)
class TestOwnerScan:
    def test_anonymous_scan_is_not_allowed(self, file: dict[str, Any]):
        """Anonymous users are not allowed to scan owners."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access(
                "files_file_scan",
                {"user": ""},
                {
                    "owner_id": file["owner_id"],
                    "owner_type": file["owner_type"],
                },
            )

    def test_authenticated_scan_is_not_allowed(self, user: dict[str, Any], file: dict[str, Any]):
        """Authenticated users are not allowed to scan owners."""
        with pytest.raises(tk.NotAuthorized):
            tk.check_access(
                "files_file_scan",
                {"user": user["name"]},
                {
                    "owner_id": file["owner_id"],
                    "owner_type": file["owner_type"],
                },
            )

    @pytest.mark.ckan_config("ckanext.files.owner.scan_as_update", False)
    def test_owner_transparent_scan(
        self,
        user: dict[str, Any],
        package_factory: types.TestFactory,
    ):
        """Scan does not work without scan-as-update because of missing auth function."""
        package = package_factory(user=user)
        with pytest.raises(ValueError):  # noqa: PT011
            tk.check_access(
                "files_file_scan",
                {"user": user["name"]},
                {
                    "owner_id": package["id"],
                    "owner_type": "package",
                },
            )

    def test_owner_scan_as_update(
        self,
        user: dict[str, Any],
        package_factory: types.TestFactory,
    ):
        """Enabled scan-as-update allows scan without additional code."""
        package = package_factory(user=user)
        tk.check_access(
            "files_file_scan",
            {"user": user["name"]},
            {
                "owner_id": package["id"],
                "owner_type": "package",
            },
        )



"""update transfer history.

Revision ID: e1e5e203d76a
Revises: 565ef5eca492
Create Date: 2025-07-29 23:14:08.426318

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "e1e5e203d76a"
down_revision = "565ef5eca492"
branch_labels = None
depends_on = None


def upgrade():
    op.alter_column("files_transfer_history", "leave_date", new_column_name="at")
    op.add_column(
        "files_transfer_history",
        sa.Column("action", sa.Text(), nullable=False, server_default="transfer"),
    )

    op.create_index("idx_file_owner_owner", "files_owner", ["owner_type", "owner_id"])


def downgrade():
    op.alter_column("files_transfer_history", "at", new_column_name="leave_date")
    op.drop_column("files_transfer_history", "action")

    op.drop_index(
        "idx_file_owner_owner",
        "files_owner",
    )



from __future__ import annotations

from collections.abc import Iterable
from typing import Any

import click
import sqlalchemy as sa

import ckan.plugins.toolkit as tk
from ckan import model
from ckan.lib.search import rebuild

from ckanext.files import shared


@click.group()
def group():
    """Migrate from original CKAN implementation."""


@group.command("groups")
@click.argument("storage_name")
def migrate_groups(storage_name: str):
    """Migrate group images to specified storage."""
    content = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        {"storage": storage_name, "rows": 0},
    )
    if not content["count"]:
        tk.error_shout(f"Storage {storage_name} contains 0 files.")
        tk.error_shout("Make sure it points to directory with group images and run:")
        tk.error_shout(f"\tckan files scan -s {storage_name} -t")
        raise click.Abort

    click.echo(f"Found {content['count']} files. Searching file owners...")

    files = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        {"storage": storage_name, "rows": content["count"]},
    )["results"]

    unowned: list[dict[str, Any]] = []
    owned: dict[tuple[str, bool], dict[str, Any]] = {}

    bar: Iterable[Any]
    with click.progressbar(files) as bar:
        for info in bar:
            group = model.Session.scalar(
                sa.select(model.Group).where(model.Group.image_url == info["location"]),
            )

            if group:
                owned[(group.id, group.is_organization)] = info
            else:
                unowned.append(info)

    click.echo(f"Located owners for {len(owned)} files out of {content['count']}.")
    if click.confirm("Show group IDs and corresponding file?"):
        for key, info in owned.items():
            click.echo(f"{key[0]}: {info['location']}")

    if unowned and click.confirm("Show files that do not belong to any group?"):
        for info in unowned:
            click.echo(f"{info['location']}")

    if click.confirm("Transfer file ownership to group identified in previous steps?"):
        with click.progressbar(owned.items()) as bar:
            for key, info in bar:
                group_id, is_organization = key
                bar.label = f"Transfering {info['location']}"
                tk.get_action("files_transfer_ownership")(
                    {"ignore_auth": True},
                    {
                        "pin": True,
                        "force": True,
                        "id": info["id"],
                        "owner_type": "organization" if is_organization else "group",
                        "owner_id": group_id,
                    },
                )


@group.command("users")
@click.argument("storage_name")
def migrate_users(storage_name: str):
    """Migrate user avatars to specified storage."""
    content = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        {"storage": storage_name, "rows": 0},
    )
    if not content["count"]:
        tk.error_shout(f"Storage {storage_name} contains 0 files.")
        tk.error_shout("Make sure it points to directory with user images and run:")
        tk.error_shout(f"\tckan files scan -s {storage_name} -t")
        raise click.Abort

    click.echo(f"Found {content['count']} files. Searching file owners...")

    files = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        {"storage": storage_name, "rows": content["count"]},
    )["results"]

    unowned: list[dict[str, Any]] = []
    owned: dict[str, dict[str, Any]] = {}

    bar: Iterable[Any]
    with click.progressbar(files) as bar:
        for info in bar:
            user = model.Session.scalar(
                sa.select(model.User).where(model.User.image_url == info["location"]),
            )

            if user:
                owned[user.id] = info

            else:
                unowned.append(info)

    click.echo(f"Located owners for {len(owned)} files out of {content['count']}.")
    if click.confirm("Show user IDs and corresponding file?"):
        for key, info in owned.items():
            click.echo(f"{key}: {info['location']}")

    if unowned and click.confirm("Show files that do not belong to any user?"):
        for info in unowned:
            click.echo(f"{info['location']}")

    if click.confirm("Transfer file ownership to users identified in previous steps?"):
        with click.progressbar(owned.items()) as bar:
            for user_id, info in bar:
                bar.label = f"Transfering {info['location']}"
                tk.get_action("files_transfer_ownership")(
                    {"ignore_auth": True},
                    {
                        "pin": True,
                        "force": True,
                        "id": info["id"],
                        "owner_type": "user",
                        "owner_id": user_id,
                    },
                )


@group.command("local-resources")
@click.argument("storage_name")
def migrate_local_resources(storage_name: str):  # noqa: C901, PLR0915, PLR0912
    """Migrate resources uploaded via original ResourceUploader."""
    content = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        {"storage": storage_name, "rows": 0},
    )
    if not content["count"]:
        tk.error_shout(f"Storage {storage_name} contains 0 files.")
        tk.error_shout("Make sure it points to directory with user images and run:")
        tk.error_shout(f"\tckan files scan -s {storage_name} -t")
        raise click.Abort

    click.echo(f"Found {content['count']} files. Searching file owners...")

    files = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        {"storage": storage_name, "rows": content["count"]},
    )["results"]

    unowned: list[dict[str, Any]] = []
    owned: dict[str, dict[str, Any]] = {}

    bar: Iterable[Any]
    with click.progressbar(files) as bar:
        for info in bar:
            resource = model.Session.scalar(
                sa.select(model.Resource).where(
                    model.Resource.id == "".join(info["location"].split("/")),
                ),
            )

            if resource:
                owned[resource.id] = info

            else:
                unowned.append(info)

    click.echo(f"Located owners for {len(owned)} files out of {content['count']}.")
    if click.confirm("Show user IDs and corresponding file?"):
        for key, info in owned.items():
            click.echo(f"{key}: {info['location']}")

    if unowned and click.confirm("Show files that do not belong to any user?"):
        for info in unowned:
            click.echo(f"{info['location']}")

    if click.confirm(
        "Transfer file ownership to resources identified in previous steps?",
    ):
        with click.progressbar(owned.items()) as bar:
            for resource_id, info in bar:
                bar.label = f"Transfering {info['location']}"

                tk.get_action("files_transfer_ownership")(
                    {"ignore_auth": True},
                    {
                        "pin": True,
                        "force": True,
                        "id": info["id"],
                        "owner_type": "resource",
                        "owner_id": resource_id,
                    },
                )
    packages: set[str] = set()
    if click.confirm("Modify resources, changing their url_type and url fields?"):
        with click.progressbar(owned.items()) as bar:
            for resource_id, info in bar:
                res = model.Resource.get(resource_id)
                fileobj = model.Session.get(shared.File, info["id"])
                if not fileobj:
                    tk.error_shout(
                        f"Filee {info['id']} disappeared during migration",
                    )
                    raise click.Abort

                if not res:
                    tk.error_shout(
                        f"Resource {resource_id} disappeared during migration",
                    )
                    raise click.Abort

                if not res.package_id:
                    tk.error_shout(f"Resource {resource_id} has no package_id")
                    raise click.Abort

                if res.url_type != "upload":
                    tk.error_shout(
                        f"Resource {resource_id} has suspicious " + f"url_type: {res.url_type}",
                    )
                    raise click.Abort

                res.url_type = "file"
                fileobj.name = res.url
                res.url = info["id"]
                packages.add(res.package_id)

    click.echo("Commit DB changes...")
    model.Session.commit()

    click.echo("Reindex modified packages...")
    for pkg in packages:
        rebuild(pkg)



import ckan.plugins.toolkit as tk
from ckan import types


def file_widget_file_create(context: types.Context, data_dict: types.DataDict) -> None:
    data_dict["storage"] = "default"
    return tk.get_action("files_file_create")(context, data_dict)


def file_widget_link_create(context: types.Context, data_dict: types.DataDict) -> None:
    data_dict["storage"] = "link"
    return tk.get_action("files_file_create")(context, data_dict)



from __future__ import annotations

import sqlalchemy as sa
from sqlalchemy.orm import Mapped

from ckan.lib.dictization import table_dictize
from ckan.types import Context

from .base import Base


class FilesOwner(Base):
    """Model with details about current owner of an item.

    Keyword Args:
        item_id (str): ID of the owned object
        item_type (str): type of the owned object
        owner_id (str): ID of the owner
        owner_type (str): Type of the owner
        pinned (bool): is ownership protected from transfer

    Example:
        ```python
        owner = Owner(
            item_id=file.id,
            item_type="file",
            owner_id=user.id,
            owner_type="user,
        )
        ```
    """

    __table__ = sa.Table(
        "files_owner",
        Base.metadata,
        sa.Column("item_id", sa.Text, primary_key=True),
        sa.Column("item_type", sa.Text, primary_key=True),
        sa.Column("owner_id", sa.Text, nullable=False),
        sa.Column("owner_type", sa.Text, nullable=False),
        sa.Column("pinned", sa.Boolean, default=False, nullable=False),
    )
    item_id: Mapped[str]
    item_type: Mapped[str]
    owner_id: Mapped[str]
    owner_type: Mapped[str]
    pinned: Mapped[bool]

    def dictize(self, context: Context):
        return table_dictize(self, context)

    def select_history(self):
        """Returns a select statement to fetch ownership history."""
        from .transfer_history import TransferHistory  # noqa: PLC0415

        return (
            sa.select(TransferHistory)
            .join(FilesOwner)
            .where(
                TransferHistory.item_id == self.item_id,
                TransferHistory.item_type == self.item_type,
            )
        )



"""Configuration readers of the extension.

This module contains functions that simplify accessing configuration option
from the CKAN config file.

It's recommended to use these functions istead of accessing config options by
name, if you want your code to be more compatible with different versions of
the extension.

"""

from __future__ import annotations

from collections import defaultdict
from typing import Any

import ckan.plugins.toolkit as tk

STORAGE_PREFIX = "ckanext.files.storage."
DEFAULT_STORAGE = "ckanext.files.default_storage"
CASCADE_ACCESS = "ckanext.files.owner.cascade_access"
TRANSFER_AS_UPDATE = "ckanext.files.owner.transfer_as_update"
SCAN_AS_UPDATE = "ckanext.files.owner.scan_as_update"
AUTHENTICATED_UPLOADS = "ckanext.files.authenticated_uploads.allow"
AUTHENTICATED_STORAGES = "ckanext.files.authenticated_uploads.storages"
USER_IMAGES_STORAGE = "ckanext.files.user_images_storage"
GROUP_IMAGES_STORAGE = "ckanext.files.group_images_storage"
RESOURCE_STORAGE = "ckanext.files.resources_storage"
ENABLE_RESOURCE_HACK = "ckanext.files.enable_resource_migration_template_patch"
INLINE_TYPES = "ckanext.files.inline_content_types"


def default_storage() -> str:
    """Default storage used for upload when no explicit storage specified."""
    return tk.config[DEFAULT_STORAGE]


def storages() -> dict[str, dict[str, Any]]:
    """Mapping of storage names to their settings."""
    storages = defaultdict(dict)  # type: dict[str, dict[str, Any]]
    prefix_len = len(STORAGE_PREFIX)
    for k, v in tk.config.items():
        if not k.startswith(STORAGE_PREFIX):
            continue

        try:
            name, option = k[prefix_len:].split(".", 1)
        except ValueError:
            continue

        storages[name][option] = v
    return storages


def cascade_access() -> dict[str, str]:
    """List of owner types that grant automatic access on owned file."""
    return tk.config[CASCADE_ACCESS]


def authenticated_uploads() -> bool:
    """Any authenticated user can upload files."""
    return tk.config[AUTHENTICATED_UPLOADS]


def transfer_as_update() -> bool:
    """Use `*_update` auth when transfering ownership."""
    return tk.config[TRANSFER_AS_UPDATE]


def scan_as_update() -> bool:
    """Use `*_update` auth when listing files of the owner."""
    return tk.config[SCAN_AS_UPDATE]


def authenticated_storages() -> list[str]:
    """Names of storages that can by used by non-sysadmins."""
    return tk.config[AUTHENTICATED_STORAGES]


def group_images_storage() -> str:
    """Storage used for group image uploads."""
    return tk.config[GROUP_IMAGES_STORAGE]


def user_images_storage() -> str:
    """Storage used for user image uploads."""
    return tk.config[USER_IMAGES_STORAGE]


def resources_storage() -> str:
    """Storage used for resource uploads."""
    return tk.config[RESOURCE_STORAGE]


def override_resource_form() -> bool:
    return tk.config[ENABLE_RESOURCE_HACK]


def inline_types() -> list[str]:
    return tk.config[INLINE_TYPES]



"""Internal utilities of the extension.

Do not use this module outside of the extension and do not import any other
internal module except for config, types and exceptions. Only independent tools
are stored here, to avoid import cycles.

"""

from __future__ import annotations

import logging
from collections.abc import Iterable
from typing import Any, Callable, TypeVar, cast

import file_keeper as fk
import jwt
from sqlalchemy.orm import Mapper

import ckan.plugins.toolkit as tk
from ckan import model
from ckan.lib.api_token import _get_algorithm, _get_secret  # pyright: ignore[reportPrivateUsage]

from ckanext.files import types

log = logging.getLogger(__name__)

T = TypeVar("T")

SAMPLE_SIZE = 1024 * 2


owner_getters = fk.Registry[Callable[[str], Any]]({})


def is_supported_type(content_type: str, supported: Iterable[str]) -> bool:
    """Check whether content_type it matches supported types."""
    maintype, subtype = content_type.split("/")
    desired = {content_type, maintype, subtype}
    return any(st in desired for st in supported)


def get_owner(owner_type: str, owner_id: str):
    if getter := owner_getters.get(owner_type):
        return getter(owner_id)

    owner_model = "group" if owner_type == "organization" else owner_type
    mappers: Iterable[Mapper[Any]]
    if tk.check_ckan_version("2.11"):
        mappers = model.registry.mappers
    else:
        mappers = cast(
            "Iterable[Mapper[Any]]",
            tk.BaseModel._sa_registry.mappers | model.User._sa_class_manager.registry.mappers,  # pyright: ignore[reportAttributeAccessIssue]
        )

    for mapper in mappers:
        cls = mapper.class_
        table = getattr(cls, "__table__", None)
        if table is None:
            table = getattr(mapper, "local_table", None)

        if table is not None and table.name == owner_model:
            return model.Session.get(cls, owner_id)

    log.warning("Unknown owner type %s with ID %s", owner_type, owner_id)
    return None


def encode_token(data: dict[str, Any]) -> str:
    return jwt.encode(data, _get_secret(encode=True), algorithm=_get_algorithm())


def decode_token(token: str) -> dict[str, Any]:
    return jwt.decode(token, _get_secret(encode=False), algorithms=[_get_algorithm()])


class ContextCache:
    """Cache for storing and retrieving values in the context."""

    key = "files_cache"
    cache: dict[str, Any]

    def __init__(self, context: types.Context):
        self.session = context.get("session", model.Session)
        self.cache = context.setdefault(self.key, {})  # pyright: ignore[reportArgumentType, reportCallIssue]

    def invalidate(self, type: str, id: str):
        """Invalidate a specific entry in the cache.

        Args:
            type (str): The type of the cached item.
            id (str): The identifier of the cached item.
        """
        self.bucket(type).pop(id, None)

    def bucket(self, type: str) -> dict[str, Any]:
        """Get or create a bucket for a specific type in the cache."""
        return self.cache.setdefault(type, {})

    def set(self, type: str, id: str, value: T) -> T:
        """Set a value in the bucket for specific id."""
        bucket = self.bucket(type)
        bucket[id] = value
        return value

    def get(self, type: str, id: str, compute: Callable[[], T]) -> T:
        """Retrieve a value from the cache or compute it if not present.

        :param type: The type of the cached item.
        :param id: The identifier of the cached item.
        :param compute: A function to compute the value if not cached.
        :returns: The cached value or the computed value.
        """
        bucket = self.bucket(type)
        if id not in bucket:
            bucket[id] = compute()

        return bucket[id]

    def get_model(self, type: str, id: str, model_class: type[T]) -> T | None:
        """Retrieve a model instance from the cache or the session."""
        return self.get(type, id, lambda: self.session.get(model_class, id))



"""drop multipart table.

Revision ID: 6e3c2aa192f6
Revises: e1e5e203d76a
Create Date: 2025-08-30 01:38:23.393215

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.postgresql import JSONB

# revision identifiers, used by Alembic.
revision = "6e3c2aa192f6"
down_revision = "e1e5e203d76a"
branch_labels = None
depends_on = None


def upgrade():
    op.create_index("idx_files_file_location_in_storage", "files_file", ["storage", "location"], unique=True)

    op.drop_table("files_multipart")


def downgrade():
    op.drop_index("idx_files_file_location_in_storage", "files_file")
    op.create_table(
        "files_multipart",
        sa.Column("id", sa.UnicodeText, primary_key=True),
        sa.Column("name", sa.UnicodeText, nullable=False),
        sa.Column("location", sa.Text, nullable=False, default=""),
        sa.Column("content_type", sa.Text, nullable=False, default=""),
        sa.Column("size", sa.Integer, nullable=False, default=0),
        sa.Column("hash", sa.Text, nullable=False, default=""),
        sa.Column("storage", sa.Text, nullable=False),
        sa.Column(
            "ctime",
            sa.DateTime(timezone=True),
            server_default=sa.func.now(),
        ),
        sa.Column("storage_data", JSONB, default=dict, server_default="{}"),
        sa.Column("plugin_data", JSONB, default=dict, server_default="{}"),
    )



from __future__ import annotations

from collections.abc import Iterable

import pytest

from ckanext.files import utils


@pytest.mark.parametrize(
    ("type", "supported", "outcome"),
    [
        ("text/csv", ["csv"], True),
        ("text/csv", ["json", "text"], True),
        ("text/csv", ["application/json", "text/plain", "text/csv", "image/png"], True),
        ("text/csv", ["json", "image"], False),
        ("text/csv", ["application/csv"], False),
        ("text/csv", ["text/plain"], False),
        ("text/csv", ["non-csv"], False),
    ],
)
def test_is_supported_type(type: str, supported: Iterable[str], outcome: bool):
    assert utils.is_supported_type(type, supported) is outcome



from __future__ import annotations

import copy
from datetime import datetime
from typing import Any, Literal

import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Mapped, foreign, relationship

from ckan.lib.dictization import table_dictize
from ckan.model.types import make_uuid

from .base import Base, now
from .owner import FilesOwner

foreign: Any


class FilesFile(Base):
    """Model with file details.

    Keyword Args:
        name (str): name shown to users
        location (str): location of the file inside storage
        content_type (str): MIMEtype
        size (int): size in bytes
        hash (str): checksum
        storage (str): storage that contains the file
        ctime (datetime): date of creation
        mtime (datetime | None): date of the last update
        atime (datetime | None): date of last access(unstable)
        storage_data (dict[str, Any]): additional data set by storage
        plugin_data (dict[str, Any]): additional data set by plugins

    Example:
        ```python
        file = File(
            name="file.txt",
            location="relative/path/safe-name.txt",
            content_type="text/plain",
            size=100,
            hash="abc123",
            storage="default",
        )
        ```
    """

    __table__ = sa.Table(
        "files_file",
        Base.metadata,
        sa.Column("id", sa.UnicodeText, primary_key=True, default=make_uuid),
        sa.Column("name", sa.UnicodeText, nullable=False),
        sa.Column("location", sa.Text, nullable=False),
        sa.Column(
            "content_type",
            sa.Text,
            nullable=False,
            default="application/octet-stream",
        ),
        sa.Column("size", sa.Integer, nullable=False, default=0),
        sa.Column("hash", sa.Text, nullable=False, default=""),
        sa.Column("storage", sa.Text, nullable=False),
        sa.Column(
            "ctime",
            sa.DateTime(timezone=True),
            nullable=False,
            default=now,
            server_default=sa.func.now(),
        ),
        sa.Column("mtime", sa.DateTime(timezone=True)),
        sa.Column("atime", sa.DateTime(timezone=True)),
        sa.Column("storage_data", JSONB, default=dict, server_default="{}"),
        sa.Column("plugin_data", JSONB, default=dict, server_default="{}"),
    )

    id: Mapped[str]

    name: Mapped[str]
    location: Mapped[str]
    content_type: Mapped[str]
    size: Mapped[int]
    hash: Mapped[str]

    storage: Mapped[str]

    ctime: Mapped[datetime]
    mtime: Mapped[datetime | None]
    atime: Mapped[datetime | None]

    storage_data: Mapped[dict[str, Any]]
    plugin_data: Mapped[dict[str, Any]]

    owner: Mapped[FilesOwner | None] = relationship(  # type: ignore
        FilesOwner,
        primaryjoin=sa.and_(
            FilesOwner.item_id == foreign(__table__.c.id),
            FilesOwner.item_type == "file",
        ),
        single_parent=True,
        uselist=False,
        cascade="delete, delete-orphan",
        lazy="joined",
    )

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        if not self.id:
            self.id = make_uuid()

    def dictize(self, context: Any) -> dict[str, Any]:
        result = table_dictize(self, context)
        result["storage_data"] = copy.deepcopy(result["storage_data"])

        if self.owner:
            result["owner_type"] = self.owner.owner_type
            result["owner_id"] = self.owner.owner_id
            result["pinned"] = self.owner.pinned

        else:
            result["owner_type"] = None
            result["owner_id"] = None
            result["pinned"] = False

        plugin_data = result.pop("plugin_data")
        if context.get("include_plugin_data"):
            result["plugin_data"] = copy.deepcopy(plugin_data)

        return result

    def touch(
        self,
        access: bool = True,
        modification: bool = False,
        moment: datetime | None = None,
    ):
        if not moment:
            moment = now()

        if access:
            self.atime = moment

        if modification:
            self.mtime = moment

    def patch_data(
        self,
        patch: dict[str, Any],
        dict_path: list[str] | None = None,
        prop: Literal["storage_data", "plugin_data"] = "plugin_data",
    ) -> dict[str, Any]:
        data: dict[str, Any] = copy.deepcopy(getattr(self, prop))

        target: Any = data
        if dict_path:
            for part in dict_path:
                target = target.setdefault(part, {})  # pyright: ignore[reportUnknownVariableType]
                if not isinstance(target, dict):
                    raise TypeError(part)

        target.update(patch)

        setattr(self, prop, data)
        return data

    @classmethod
    def by_location(cls, location: str, storage: str):
        return sa.select(cls).where(
            cls.location == location,
            cls.storage == storage,
        )



from __future__ import annotations

import click
import file_keeper as fk
import sqlalchemy as sa

import ckan.plugins.toolkit as tk
from ckan import model

from ckanext.files import shared


@click.group()
def group():
    """Storage maintenance."""


storage_option = click.option(
    "-s",
    "--storage-name",
    help="Name of the configured storage",
)


@group.command()
@storage_option
@click.option("--remove", is_flag=True, help="Remove files")
def empty_owner(storage_name: str | None, remove: bool):
    """Manage files that have no owner."""
    storage_name = storage_name or shared.config.default_storage()
    try:
        storage = shared.get_storage(storage_name)
    except shared.exc.UnknownStorageError as err:
        tk.error_shout(err)
        raise click.Abort from err

    if remove and not storage.supports(shared.Capability.REMOVE):
        tk.error_shout(f"Storage {storage_name} does not support file removal")
        raise click.Abort

    stmt = (
        sa.select(shared.File)
        .outerjoin(shared.File.owner)
        .where(shared.File.storage == storage_name, shared.Owner.owner_id.is_(None))
    )

    total = model.Session.scalar(stmt.with_only_columns(sa.func.count()))
    if not total:
        click.echo(f"Every file in storage {storage_name} has owner reference")
        return
    click.echo("Following files do not have owner reference")

    for file in model.Session.scalars(stmt):
        size = fk.humanize_filesize(file.size)
        click.echo(f"\t{file.id}: {file.name} [{file.content_type}, {size}]")

    if remove and click.confirm("Do you want to delete these files?"):
        action = tk.get_action("files_file_delete")

        with click.progressbar(model.Session.scalars(stmt), length=total) as bar:
            for file in bar:
                action({"ignore_auth": True}, {"id": file.id})


@group.command()
@storage_option
@click.option("--remove", is_flag=True, help="Remove files")
def invalid_owner(storage_name: str | None, remove: bool):
    """Manage files that has suspicious owner reference."""
    storage_name = storage_name or shared.config.default_storage()
    try:
        storage = shared.get_storage(storage_name)
    except shared.exc.UnknownStorageError as err:
        tk.error_shout(err)
        raise click.Abort from err

    if remove and not storage.supports(shared.Capability.REMOVE):
        tk.error_shout(f"Storage {storage_name} does not support file removal")
        raise click.Abort

    stmt = sa.select(shared.File).join(shared.File.owner).where(shared.File.storage == storage_name)

    files = [f for f in model.Session.scalars(stmt) if f.owner is None]

    if not files:
        click.echo(
            f"Every owned file in storage {storage_name} has valid owner reference",
        )
        return

    click.echo("Following files have dangling owner reference")
    for file in files:
        size = fk.humanize_filesize(file.size)
        if owner := file.owner:
            click.echo(
                f"\t{file.id}: {file.name} [{file.content_type}, {size}]. "
                + f"Owner: {owner.owner_type} {owner.owner_id}",
            )

    if remove and click.confirm("Do you want to delete these files?"):
        action = tk.get_action("files_file_delete")

        with click.progressbar(files) as bar:
            for file in bar:
                action({"ignore_auth": True}, {"id": file.id})


@group.command()
@storage_option
@click.option("--remove", is_flag=True, help="Remove files")
def missing_files(storage_name: str | None, remove: bool):
    """Manage files that do not exist in storage."""
    storage_name = storage_name or shared.config.default_storage()
    try:
        storage = shared.get_storage(storage_name)
    except shared.exc.UnknownStorageError as err:
        tk.error_shout(err)
        raise click.Abort from err

    if not storage.supports(shared.Capability.EXISTS):
        tk.error_shout(
            f"Storage {storage_name} does not support file availability checks",
        )
        raise click.Abort

    if remove and not storage.supports(shared.Capability.REMOVE):
        tk.error_shout(f"Storage {storage_name} does not support file removal")
        raise click.Abort

    stmt = sa.select(shared.File).where(shared.File.storage == storage_name)
    total = model.Session.scalar(stmt.with_only_columns(sa.func.count()))
    missing: list[shared.File] = []
    with click.progressbar(model.Session.scalars(stmt), length=total) as bar:
        for file in bar:
            data = shared.FileData.from_object(file)
            if not storage.exists(data):
                missing.append(file)

    if not missing:
        click.echo(
            f"No missing files located in storage {storage_name}",
        )
        return

    click.echo(f"Following files are not found in the storage {storage_name}")
    for file in missing:
        size = fk.humanize_filesize(file.size)
        click.echo(
            f"\t{file.id}: {file.name} [{file.content_type}, {size}]",
        )

    if remove and click.confirm("Do you want to delete these files from registry?"):
        action = tk.get_action("files_file_delete")

        with click.progressbar(missing) as bar:
            for file in bar:
                action({"ignore_auth": True}, {"id": file.id})



"""create transfer history table.

Revision ID: 6a6ea2472155
Revises: a11926fa799e
Create Date: 2024-06-16 16:23:27.102126

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "6a6ea2472155"
down_revision = "a11926fa799e"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "files_transfer_history",
        sa.Column("id", sa.Text, primary_key=True),
        sa.Column("item_id", sa.Text, nullable=False),
        sa.Column("item_type", sa.Text, nullable=False),
        sa.Column("owner_id", sa.Text, nullable=False),
        sa.Column("owner_type", sa.Text, nullable=False),
        sa.Column(
            "leave_date",
            sa.DateTime(timezone=True),
            server_default=sa.func.now(),
            nullable=False,
        ),
        sa.Column("actor", sa.Text, nullable=False),
        sa.Index("idx_item", "item_id", "item_type"),
        sa.ForeignKeyConstraint(
            ["item_id", "item_type"],
            ["files_owner.item_id", "files_owner.item_type"],
        ),
    )


def downgrade():
    op.drop_table("files_transfer_history")



from __future__ import annotations

import logging
from functools import partial
from typing import Any

import file_keeper as fk
import jwt
from flask import Blueprint, jsonify

import ckan.plugins.toolkit as tk
from ckan import model
from ckan.common import streaming_response
from ckan.lib.pagination import Page
from ckan.logic import parse_params
from ckan.types import Response
from ckan.views.resource import download

from ckanext.files import shared, utils

log = logging.getLogger(__name__)
bp = Blueprint("files", __name__)

__all__ = ["bp"]


def not_found_handler(error: tk.ObjectNotFound) -> tuple[str, int]:
    """Generic handler for ObjectNotFound exception."""
    return (
        tk.render(
            "error_document_template.html",
            {
                "code": 404,
                "content": f"Object not found: {error.message}",
                "name": "Not found",
            },
        ),
        404,
    )


def not_authorized_handler(error: tk.NotAuthorized) -> tuple[str, int]:
    """Generic handler for NotAuthorized exception."""
    return (
        tk.render(
            "error_document_template.html",
            {
                "code": 403,
                "content": error.message or "Not authorized to view this page",
                "name": "Not authorized",
            },
        ),
        403,
    )


bp.register_error_handler(tk.ObjectNotFound, not_found_handler)
bp.register_error_handler(tk.NotAuthorized, not_authorized_handler)


def get_blueprints():
    return [bp]


@bp.route("/file/download/<file_id>")
def dispatch_download(file_id: str) -> Response:
    tk.check_access("files_permission_download_file", {}, {"id": file_id})
    item = model.Session.get(shared.File, file_id)
    if not item:
        raise tk.ObjectNotFound("file")

    storage = shared.get_storage(item.storage)
    data = shared.FileData.from_object(item)

    if isinstance(storage, shared.Storage):
        return storage.as_response(data)

    link = storage.permanent_link(data)

    if link:
        return tk.redirect_to(link)

    if resp := _streaming_file(item, storage, data):
        return resp

    return tk.abort(422, "File is not downloadable")


def _streaming_file(
    item: shared.File,
    storage: fk.Storage,
    data: shared.FileData,
) -> Response | None:
    if storage.supports(shared.Capability.STREAM):
        resp = streaming_response(storage.stream(data), data.content_type)
        if utils.is_supported_type(item.content_type, shared.config.inline_types()):
            resp.headers["content-disposition"] = f"inline; filename={item.name}"
        else:
            resp.headers["content-disposition"] = f"attachment; filename={item.name}"

        item.touch()
        model.Session.commit()

        return resp


@bp.route("/file/token-download/<token>")
def temporal_download(token: str) -> Response:
    try:
        data = utils.decode_token(token)

    except jwt.ExpiredSignatureError as err:
        log.debug("Expired file-download token: %s", err)
        raise tk.ObjectNotFound("file") from err

    except jwt.InvalidTokenError as err:
        log.debug("Cannot decode file-download token: %s", err)
        raise tk.ObjectNotFound("file") from err

    if data.get("topic") != "download_file":
        raise tk.ObjectNotFound("file")

    if "id" in data:
        item = model.Session.get(shared.File, data["sub"])
    elif "location" in data and "storage" in data:
        item = model.Session.scalar(shared.File.by_location(data["location"], data["storage"]))
    else:
        item = None

    if not item:
        raise tk.ObjectNotFound("file")

    storage = shared.get_storage(item.storage)
    data = shared.FileData.from_object(item)

    if isinstance(storage, shared.Storage):
        return storage.as_response(data)

    if resp := _streaming_file(item, storage, data):
        return resp

    return tk.abort(422, "File is not downloadable")


def _pager_url(*args: Any, **kwargs: Any) -> str:
    """Generic URL builder for `url` parameter of ckan.lib.pagination.Page.

    It generates pagination link keeping all the parameters from URL and query
    string.
    """
    params = {k: v for k, v in tk.request.args.items() if k != "page"}
    view_args: dict[str, Any] = tk.request.view_args or {}
    params.update(
        dict(view_args.items()),
    )
    params.update(kwargs)
    return tk.h.pager_url(*args, **params)


@bp.route("/<owner_type>/<owner_id>/files/list")
def list_files(owner_type: str, owner_id: str) -> str:
    owner = utils.get_owner(owner_type, owner_id)
    if owner and hasattr(owner, "id"):
        owner_id = owner.id

    rows = 10
    params = tk.request.args
    page = tk.h.get_page_number(params)
    start = rows * page - rows

    search_dict: dict[str, Any] = {
        "rows": rows,
        "start": start,
        "owner_type": owner_type,
        "owner_id": owner_id,
        "sort": params.get("sort", "ctime"),
        "reverse": params.get("reverse", True),
    }

    if "storage" in params:
        search_dict["storage"] = params["storage"]

    result: dict[str, Any]
    try:
        result = tk.get_action("files_file_scan")({}, search_dict)
    except tk.ValidationError as err:
        for k, v in err.error_summary.items():
            tk.h.flash_error(f"{k}: {v}")

        result = {"count": 0, "results": []}

    pager = Page(
        result["results"],
        items_per_page=rows,
        page=page,
        item_count=result["count"],
        url=partial(_pager_url),
    )

    tpl_data: dict[str, Any] = {
        "owner": owner,
        "pager": pager,
        "owner_type": owner_type,
        "owner_id": owner_id,
    }

    return tk.render("files/index.html", tpl_data)


@bp.route("/<owner_type>/<owner_id>/files/delete/<file_id>", methods=["POST", "GET"])
def delete_file(
    owner_type: str,
    owner_id: str,
    file_id: str,
) -> str | shared.types.Response:
    if tk.request.method == "POST":
        try:
            tk.get_action("files_file_delete")({}, {"id": file_id})
        except tk.NotAuthorized as err:
            tk.h.flash_error(err)
        else:
            came_from = tk.h.get_request_param("came_from")
            if came_from:
                return tk.redirect_to(came_from)

            return tk.redirect_to(
                "files.list_files",
                owner_type=owner_type,
                owner_id=owner_id,
            )

    tk.check_access("files_file_delete", {}, {"id": file_id})
    info = tk.get_action("files_file_show")({}, {"id": file_id})
    return tk.render("files/delete.html", {"file": info})


@bp.route("/api/util/files_autocomplete_own_files")
def autocomplete_own_files() -> Any:
    tk.check_access("files_autocomplete_available_resource_files", {}, {})

    params: dict[str, Any] = parse_params(tk.request.args)

    q = params.pop("q", "")

    params.update(
        {
            "owner_type": "user",
            "owner_id": tk.current_user.is_authenticated and tk.current_user.id,  # type: ignore
            "pinned": False,
            "name": ["like", f"%{q}%"],
        },
    )

    result = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        params,
    )

    return jsonify(
        {
            "ResultSet": {
                "Result": [
                    dict(
                        item,
                        label="{name} [{content_type}, {size}]".format(
                            name=item["name"],
                            content_type=tk.h.unified_resource_format(
                                item["content_type"],
                            ),
                            size=fk.humanize_filesize(item["size"]),
                        ),
                    )
                    for item in result["results"]
                ],
            },
        },
    )


@bp.route("/api/util/files_autocomplete_available_resource_files")
def autocomplete_available_resource_files() -> Any:
    tk.check_access("files_autocomplete_available_resource_files", {}, {})
    q = tk.request.args.get("incomplete")

    result = tk.get_action("files_file_search")(
        {"ignore_auth": True},
        {
            "owner_type": "user",
            "owner_id": tk.current_user.is_authenticated and tk.current_user.id,
            "pinned": False,
            "storage": shared.config.resources_storage(),
            "name": ["like", f"%{q}%"],
        },
    )

    return jsonify(
        {
            "ResultSet": {
                "Result": [
                    {
                        "id": item["id"],
                        "name": item["name"],
                        "label": "{name} [{content_type}, {size}]".format(
                            name=item["name"],
                            content_type=tk.h.unified_resource_format(
                                item["content_type"],
                            ),
                            size=fk.humanize_filesize(item["size"]),
                        ),
                    }
                    for item in result["results"]
                ],
            },
        },
    )


@bp.route(
    "/<package_type>/<id>/resource/<resource_id>/download",
    defaults={"package_type": "dataset"},
)
@bp.route(
    "/<package_type>/<id>/<resource_id>/download/<filename>",
    defaults={"package_type": "dataset"},
)
def resource_download(
    package_type: str,
    id: str,
    resource_id: str,
    filename: str | None = None,
):
    try:
        resource = tk.get_action("resource_show")({}, {"id": resource_id})
    except tk.ObjectNotFound:
        return tk.abort(404, tk._("Resource not found"))
    except tk.NotAuthorized:
        return tk.abort(403, tk._("Not authorized to download resource"))

    if resource.get("url_type") != "file":
        return download(package_type, id, resource_id, filename)

    file_id = resource["url"].rsplit("/", 1)[-1]
    return dispatch_download(file_id)



from __future__ import annotations

import dataclasses

from file_keeper.default.adapters import filebin
from typing_extensions import override

from ckan.config.declaration import Declaration, Key

from ckanext.files import shared


@dataclasses.dataclass()
class Settings(shared.Settings, filebin.Settings):
    pass


class FilebinStorage(shared.Storage, filebin.FilebinStorage):  # pyright: ignore[reportIncompatibleVariableOverride]
    hidden = True

    settings: Settings  # pyright: ignore[reportIncompatibleVariableOverride]
    SettingsFactory = Settings
    ReaderFactory = type("Reader", (shared.Reader, filebin.Reader), {})

    @override
    @classmethod
    def declare_config_options(cls, declaration: Declaration, key: Key):
        super().declare_config_options(declaration, key)

        declaration.declare(key.bin).required().set_description("ID of the bin")
        declaration.declare_int(key.timeout, 10).set_description(
            "Timeout of requests to filebin.net",
        )
